#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/driver.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca

from argparse import ArgumentParser
import os.path as op
import tempfile
import json
import sys
import os

from clowdr.controller import metadata, launcher
from clowdr.task import processTask
from clowdr.server import shareapp, updateIndex
from clowdr import utils


def local(tool, invocation, clowdrloc, dataloc, **kwargs):
    """local
    Launches a pipeline locally through the Clowdr wrappers.

    Parameters
    ----------
    tool : str
        Path to a boutiques descriptor for the tool to be run
    invocation : str
        Path to a boutiques invocation for the tool and parameters to be run
    clowdrloc : str
        Path for storing Clowdr intermediate files and outputs
    dataloc : str
        Path for accessing input data. If local, provide the hostname and
        optionally a path. If on S3, provide an S3 path.
    **kwargs : dict
        Arbitrary keyword arguments. Currently supported arguments:
        - verbose : bool
            Toggle verbose output printing
        - dev : bool
            Toggle dev mode (only runs first execution in the specified set)

        Additionally, transfers all keyword arguments accepted by "processTask"

    Returns
    -------
    int
        The exit-code returned by the task being executed
    """
    # TODO: scrub inputs
    [tasks, invocs] = metadata.consolidateTask(tool, invocation, clowdrloc,
                                               dataloc, **kwargs)
    if kwargs.get("dev"):
        tasks = [tasks[0]]  # Just launch the first task in dev

    taskdir = op.dirname(utils.truepath(tasks[0]))
    os.chdir(taskdir)
    for task in tasks:
        processTask(task, taskdir, local=True, **kwargs)

    if kwargs.get("verbose"):
        print(taskdir)
    return taskdir


def cluster(tool, invocation, clowdrloc, dataloc, cluster, **kwargs):
    """cluster
    Launches a pipeline locally through the Clowdr wrappers.

    Parameters
    ----------
    tool : str
        Path to a boutiques descriptor for the tool to be run
    invocation : str
        Path to a boutiques invocation for the tool and parameters to be run
    clowdrloc : str
        Path for storing Clowdr intermediate files and outputs
    dataloc : str
        Path for accessing input data. If local, provide the hostname and
        optionally a path. If on S3, provide an S3 path.
    cluster : str
        Scheduler on the cluster being used. Currently, the only supported mode
        is slurm.
    **kwargs : dict
        Arbitrary keyword arguments. Currently supported arguments:
        - account : str
            Account for the cluster scheduler
        - jobname : str
            Base-name for the jobs as they will appear in the scheduler
        - verbose : bool
            Toggle verbose output printing
        - dev : bool
            Toggle dev mode (only runs first execution in the specified set)

        Additionally, transfers all keyword arguments accepted by both of
        "controller.metadata.consolidateTask" and "task.processTask"

    Returns
    -------
    int
        The exit-code returned by the task being executed
    """
    # TODO: scrub inputs
    tool = utils.truepath(tool)
    if kwargs.get("simg"):
        kwargs["simg"] = utils.truepath(kwargs["simg"])


    from slurmpy import Slurm

    if kwargs.get("verbose"):
        print("Consolidating metadata...")
    [tasks, invocs] = metadata.consolidateTask(tool, invocation, clowdrloc,
                                               dataloc, **kwargs)
    if kwargs.get("dev"):
        tasks = [tasks[0]]  # Just launch the first task in dev

    taskdir = op.dirname(utils.truepath(tasks[0]))
    try:
        os.mkdir(taskdir)
    except FileExistsError:
        pass
    os.chdir(taskdir)

    with open(tool) as fhandle:
        container = json.load(fhandle).get("container-image")
    if container:
        if kwargs.get("verbose"):
            print("Getting container...")
        outp = utils.getContainer(taskdir, container, **kwargs)
        if kwargs.get("verbose"):
            print(outp)

    jobname = kwargs.get("jobname") if kwargs.get("jobname") else "clowdrtask"
    slurm_args = {}
    if kwargs.get("slurm_args"):
        for opt in kwargs.get("slurm_args").split(","):
            k, v = opt.split(":")[0], opt.split(":")[1:]
            v = ":".join(v)
            slurm_args[k] = v
    job = Slurm(jobname, slurm_args)

    script = "clowdr run {} -c {} --local"
    if kwargs.get("workdir"):
        script += " -w {}".format(kwargs["workdir"])
    if kwargs.get("volumes"):
        script += " ".join([" -v {}".format(vol)
                            for vol in kwargs.get("volumes")])

    for task in tasks:
        job.run(script.format(task, taskdir))

    if kwargs.get("verbose"):
        print(taskdir)
    return taskdir


def cloud(tool, invocation, clowdrloc, dataloc, endpoint, auth, **kwargs):
    """cloud
    Launches a pipeline locally at scale through Clowdr.

    Parameters
    ----------
    tool : str
        Path to a boutiques descriptor for the tool to be run
    invocation : str
        Path to a boutiques invocation for the tool and parameters to be run
    clowdrloc : str
        Path on S3 for storing Clowdr intermediate files and outputs
    dataloc : str
        Path on S3 for accessing input data
    endpoint : str
        Which endpoint to use for deployment
    auth : str
        Credentials for Amazon with access to dataloc, clowdrloc, and Batch
    **kwargs : dict
        Arbitrary keyword arguments (i.e. {'verbose': True})

    Returns
    -------
    int
        The exit-code returned by the task being executed
    """
    # TODO: scrub inputs better
    clowdrloc = clowdrloc.strip('/')

    # Create temp dir for clowdrloc 
    tmploc = utils.truepath(tempfile.mkdtemp())

    [tasks, invocs] = metadata.consolidateTask(tool, invocation, tmploc,
                                               dataloc, **kwargs)
    metadata.prepareForRemote(tasks, tmploc, clowdrloc)
    tasks_remote = [task for task in utils.post(tmploc, clowdrloc)
                    if "task-" in task]

    resource = launcher.configureResource(endpoint, auth, **kwargs)
    jids = []
    for task in tasks_remote:
        jids += [resource.launchJob(task)]

    taskdir = op.dirname(utils.truepath(tasks_remote[0]))
    print(taskdir)
    return taskdir, jids


def share(clowdrloc, **kwargs):
    """share
    Launches a simple web server which showcases all runs at the clowdrloc.

    Parameters
    ----------
    clowdrloc : str
        Path with Clowdr metdata files (returned from "local" and "deploy")
    **kwargs : dict
        Arbitrary keyword arguments (i.e. {'verbose': True})

    Returns
    -------
    None
    """
    # TODO: scrub inputs
    shareapp.config["clowdrloc"] = clowdrloc
    shareapp.config["tmpdir"] = tempfile.mkdtemp()

    updateIndex()

    host = kwargs["host"] if kwargs.get("host") else "0.0.0.0"
    shareapp.run(host=host, debug=kwargs.get("debug"))


def main(args=None):
    """main
    Command-line API wrapper for Clowdr as a CLI, not Python API.

    Parameters
    ----------
    args: list
        List of all command-line arguments being passed

    Returns
    -------
    int
        The exit-code returned by the task being executed
    """
    desc = "Interface for launching Boutiques task locally and in the cloud"
    parser = ArgumentParser("Clowdr CLI", description=desc)
    subparsers = parser.add_subparsers(help="Modes of operation", dest="mode")

    # Local Parser
    parser_loc = subparsers.add_parser("local")
    parser_loc.add_argument("tool", help="boutiques descriptor for a tool")
    parser_loc.add_argument("invocation", help="input(s) for the tool")
    parser_loc.add_argument("clowdrloc", help="location locally for clowdr")
    parser_loc.add_argument("dataloc", help="location locally or s3 for data")
    parser_loc.add_argument("--workdir", "-w", action="store")
    parser_loc.add_argument("--volumes", "-v", action="append")
    parser_loc.add_argument("--user", "-u", action="store_true")

    parser_loc.add_argument("--verbose", "-V", action="store_true")
    parser_loc.add_argument("--bids", "-b", action="store_true")
    parser_loc.add_argument("--dev", "-d", action="store_true")

    parser_loc.set_defaults(func=local)

    # Cluster Parser
    parser_cls = subparsers.add_parser("cluster")
    parser_cls.add_argument("tool", help="boutiques descriptor for a tool")
    parser_cls.add_argument("invocation", help="input(s) for the tool")
    parser_cls.add_argument("clowdrloc", help="location locally for clowdr")
    parser_cls.add_argument("dataloc", help="location locally or s3 for data")
    parser_cls.add_argument("cluster", help="cluster type", choices=["slurm"])
    parser_cls.add_argument("--jobname", "-n", action="store")
    parser_cls.add_argument("--slurm_args", action="store")
    parser_cls.add_argument("--workdir", "-w", action="store")
    parser_cls.add_argument("--volumes", "-v", action="append")
    parser_cls.add_argument("--simg", "-s", action="store")

    parser_cls.add_argument("--verbose", "-V", action="store_true")
    parser_cls.add_argument("--bids", "-b", action="store_true")
    parser_cls.add_argument("--dev", "-d", action="store_true")

    parser_cls.set_defaults(func=cluster)

    # Cloud Parser
    parser_cld = subparsers.add_parser("cloud")
    parser_cld.add_argument("tool",  help="boutiques descriptor for a tool")
    parser_cld.add_argument("invocation", help="input(s) for the tool")
    parser_cld.add_argument("clowdrloc", help="location on s3 for clowdr")
    parser_cld.add_argument("dataloc", help="location on s3 of data")
    parser_cld.add_argument("cloud", help="cloud endpoint", choices=["aws"])
    parser_cld.add_argument("auth", help="credentials for the remote resource")
    parser_cld.add_argument("--region", "-r", action="store")

    parser_cld.add_argument("--verbose", "-V", action="store_true")
    parser_cld.add_argument("--bids", "-b", action="store_true")
    parser_cld.add_argument("--dev", "-d", action="store_true")

    parser_cld.set_defaults(func=cloud)

    # Share Parser
    parser_shr = subparsers.add_parser("share")
    parser_shr.add_argument("clowdrloc", help="local or s3 location for clowdr")
    parser_shr.add_argument("--debug", "-d", action="store_true")

    parser_shr.set_defaults(func=share)

    # Run Parser
    parser_run = subparsers.add_parser("run")
    parser_run.add_argument("metadata", help="task metadata file")
    parser_run.add_argument("--clowdrloc", "-c", action="store",
                            help="task output directory")
    parser_run.add_argument("--local", "-l", action="store_true")
    parser_run.add_argument("--workdir", "-w", action="store")
    parser_run.add_argument("--volumes", "-v", action="append")

    parser_run.add_argument("--verbose", "-V", action="store_true")

    parser_run.set_defaults(func=processTask)

    # Parse arguments
    inps = parser.parse_args(args) if args is not None else parser.parse_args()
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()
    else:
        inps.func(**vars(inps))


if __name__ == "__main__":
    main()




#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/share/server.py
# Created by Greg Kiar on 2018-03-01.
# Email: gkiar@mcin.ca

from flask import Flask, render_template, redirect
import os.path as op
import datetime
import tempfile
import boto3
import json
import sys
import re
import os
from botocore.handlers import disable_signing

from clowdr import utils


shareapp = Flask(__name__)

@shareapp.route("/")
def index():
    with open(shareapp.config.get("datapath")) as fhandle:
        data = json.load(fhandle)
    return render_template("index.html", data=data)


@shareapp.route("/refresh")
def update():
    updateIndex()
    return redirect("/")


def parseJSON(outdir, objlist, s3bool=True, **kwargs):
    tmplist = []
    for obj in objlist:
        tmpdict = {}
        key = obj["key"]
        for key2 in obj.keys():
            tmpdict[key2] = obj[key2]

        with open(obj["fname"]) as fhandle:
            tmpdict["contents"] = json.load(fhandle)

        if kwargs.get("descriptor"):
            tmpdict["name"] = tmpdict["contents"]["name"]
        elif kwargs.get("invocation"):
            tmpdict["name"] = op.basename(key)
        elif kwargs.get("summary"):
            tmpdict["id"] = op.splitext(op.basename(obj["fname"]))[0].split('-')[1]
            outfname = op.basename(tmpdict["contents"]["stdout"])
            with open(op.join(outdir, outfname)) as fhandle:
                tmpdict["out"] = fhandle.read()
            errfname = op.basename(tmpdict["contents"]["stderr"])
            with open(op.join(outdir, errfname)) as fhandle:
                tmpdict["err"] = fhandle.read()
            tmpdict["exitcode"] = tmpdict["contents"]["exitcode"]
            tmpdur = tmpdict["contents"]["duration"]
            tmpdict["duration"] = str(datetime.timedelta(seconds=tmpdur))
            tmpdict["outputs"] = tmpdict["contents"]["outputs"]
        elif kwargs.get("task"):
            tmpdict["id"] = op.splitext(op.basename(obj["fname"]))[0].split('-')[1]
            if kwargs.get("data").get("invocation"):
                invocs = kwargs["data"]["invocation"]
                invname = op.basename(tmpdict["contents"]["invocation"])
                invoc = [inv for inv in invocs if inv["name"] == invname][0]
                tmpdict["invoc"] = invoc
            if kwargs.get("data").get("summary"):
                summars = kwargs["data"]["summary"]
                summar = [summ for summ in summars if summ["id"] == tmpdict["id"]]
                if len(summar) > 0:
                    tmpdict["summary"] = summar[0]
                else:
                    tmpdict["summary"] = {"exitcode": None}

        tmplist.append(tmpdict)
    return tmplist


def getRecords(clowdrloc, outdir, **kwargs):
    # get remote records, open local records
    try:
        bucket, rpath = utils.splitS3Path(clowdrloc)
        s3bool = True
    except AttributeError:
        s3bool = False
        hostname = kwargs.get("hostname")

    if s3bool:
        s3 = boto3.resource("s3") # , config=Config(signature_version=UNSIGNED))
        # s3.meta.client.meta.events.register('choose-signer.s3.*', disable_signing)
        cli = boto3.client("s3") # , config=Config(signature_version=UNSIGNED))
        buck = s3.Bucket(bucket)
        objs = buck.objects.filter(Prefix=rpath)
        objs = [{"key": obj.key,
                 "bucket_name": obj.bucket_name,
                 "date": obj.last_modified.strftime("%b %d, %Y (%T)"),
                 "url": cli.generate_presigned_url("get_object",
                                                   Params={"Bucket": obj.bucket_name,
                                                           "Key": obj.key},
                                                   ExpiresIn=None)}
                for obj in objs]
        for idx, obj in enumerate(objs):
            outfname = op.join(outdir, op.basename(obj["key"]))
            buck.download_file(obj["key"], Filename=outfname)
            objs[idx]["fname"] = outfname
    else:
        tsp = datetime.datetime.fromtimestamp
        objs = [{"key": op.join(clowdrloc, f),
                 "hostname": hostname,
                 "date": tsp(op.getmtime(op.join(clowdrloc, f))).strftime("%b %d, %Y (%T)"),
                 "fname": op.join(clowdrloc, f),
                 "url": None}
                for f in os.listdir(clowdrloc)]

    task = [obj for obj in objs if "task-" in obj["key"]]
    summ = [obj for obj in objs if "summary-" in obj["key"]]
    desc = [obj for obj in objs if "descriptor" in obj["key"]]
    invo = [obj for obj in objs if "invocation" in obj["key"]]
    outs = [obj for obj in objs if ".txt" in obj["key"]]
    return (s3bool, task, summ, desc, invo, outs)


def updateIndex(**kwargs):
    clowdrloc = shareapp.config.get("clowdrloc")
    tmpdir = shareapp.config.get("tmpdir")
    s3bool, task, summ, desc, invo, outs = getRecords(clowdrloc, tmpdir)

    if not s3bool:
        tmpdir = clowdrloc

    desc = parseJSON(tmpdir, desc, s3bool, descriptor=True)[0]
    invo = parseJSON(tmpdir, invo, s3bool, invocation=True)
    summ = parseJSON(tmpdir, summ, s3bool, summary=True,
                     data={"outs": outs})
    task = parseJSON(tmpdir, task, s3bool, task=True,
                     data={"invocation": invo,
                           "summary"   : summ})

    data = {"clowdrloc" : clowdrloc,
            "tool"      : desc,
            "tasks"     : task}

    fname = op.join(tmpdir, "clowdrsitedata.json")
    with open(fname, "w") as fhandle:
        fhandle.write(json.dumps(data, indent=2))

    shareapp.config["datapath"] = fname



#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/task.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca

from argparse import ArgumentParser
import os.path as op
import time
import json
import os
import re

import boutiques as bosh
from clowdr import utils


def processTask(metadata, clowdrloc=None, verbose=False, **kwargs):
    # Get metadata
    if clowdrloc is None:
        localtaskdir = "/clowtask/"
    else:
        localtaskdir = clowdrloc

    localtaskdir = op.join(localtaskdir, "clowtask_"+utils.randstring(3))
    if not op.exists(localtaskdir):
        os.makedirs(localtaskdir)

    if(verbose):
        print("Fetching metadata...")
    remotetaskdir = op.dirname(metadata)
    metadata = utils.get(metadata, localtaskdir)[0]
    task_id = metadata.split('.')[0].split('-')[-1]
    # The above grabs an ID from the form: fname-#.ext

    # Parse metadata
    metadata   = json.load(open(metadata))
    descriptor = metadata['tool']
    invocation = metadata['invocation']
    input_data = metadata['dataloc']
    output_loc = utils.truepath(metadata['taskloc'])

    if(verbose):
        print("Fetching descriptor and invocation...")
    # Get descriptor and invocation
    desc_local = utils.get(descriptor, localtaskdir)[0]
    invo_local = utils.get(invocation, localtaskdir)[0]

    # Get input data, if running remotely
    if not kwargs.get("local") and \
       any([dl.startswith("s3://") for dl in input_data]):
        if(verbose):
            print("Fetching input data...")
        localdatadir = op.join(localtaskdir, "data")
        for dataloc in input_data:
            utils.get(dataloc, localdatadir)
        # Move to correct location
        os.chdir(localdatadir)
    else:
        if(verbose):
            print("Skipping data fetch (local execution)...")
        if kwargs.get("workdir") and op.exists(kwargs.get("workdir")):
            os.chdir(kwargs["workdir"])

    if(verbose):
        print("Beginning execution...")
    # Launch task
    start_time = time.time()
    copts = ['launch', desc_local, invo_local]
    if kwargs.get("volumes"):
        volumes = " ".join(kwargs.get("volumes"))
        copts += ['-v', volumes]
    if kwargs.get("user"):
        copts += ['-u']

    bosh_output = bosh.execute(*copts)
    if(verbose):
        print(bosh_output)
    duration = time.time() - start_time

    # Get list of bosh exec outputs
    with open(desc_local) as fhandle:
        outputs_all = json.load(fhandle)["output-files"]

    outputs_present = []
    outputs_all = bosh.evaluate(desc_local, invo_local, 'output-files/')
    for outfile in outputs_all.values():
        outputs_present += [outfile] if op.exists(outfile) else []

    # Write stdout to file
    stdoutf = "stdout-{}.txt".format(task_id)
    with open(op.join(localtaskdir, stdoutf), "w") as fhandle:
        fhandle.write(bosh_output.stdout)
    utils.post(op.join(localtaskdir, stdoutf), remotetaskdir)

    # Write sterr to file
    stderrf = "stderr-{}.txt".format(task_id)
    with open(op.join(localtaskdir, stderrf), "w") as fhandle:
        fhandle.write(bosh_output.stderr)
    utils.post(op.join(localtaskdir, stderrf), remotetaskdir)

    # Write summary values to file, including:
    summary = {"duration": duration,
               "exitcode": bosh_output.exit_code,
               "outputs": [],
               "stdout": op.join(remotetaskdir, stdoutf),
               "stderr": op.join(remotetaskdir, stderrf)}

    if not kwargs.get("local"):
        if(verbose):
            print("Uploading outputs...")
        # Push outputs
        for local_output in outputs_present:
            if(verbose):
                print("{} --> {}".format(local_output, output_loc))
            summary["outputs"] += utils.post(local_output, output_loc)
    else:
        if(verbose):
            print("Skipping uploading outputs (local execution)...")
        summary["outputs"] = outputs_present

    summarf = "summary-{}.json".format(task_id)
    with open(op.join(localtaskdir, summarf), "w") as fhandle:
        fhandle.write(json.dumps(summary) + "\n")
    utils.post(op.join(localtaskdir, summarf), remotetaskdir)




from __future__ import absolute_import

from .task import processTask
from .controller import metadata, launcher
from .endpoint import AWS, remote
from .driver import local, cluster, cloud, share

__all__ = ['task', 'controller', 'driver', 'endpoint']

